
%% author: Alvaro Feito, 14 July 2009
%% Imperial College London
%% based on code from Kenneth Pregnell, Martin Plenio, Alvaro Feito
%% used in arXiv:0807.2444 and arXiv:0906.3440

clear all

%%
%%%  DETECTOR TOMOGRAPHY - Semi Definite Program for Reconstruction %%%%%%



%%% The aim is to reconstruct the POVMs (POV) from an unknown detector.  
%%% To do so a family of well known states (rho) is measured with probability:
%%% ---   Pr = Tr( rho POV)  ---
%%% This code imports the data "Pr", defines the probe state "rho" and
%%% reconstructs the unknown POVM elements POV.


%%The code has three main parts:

% (I)   Import the Data
% (II)  Define the Input state
% (III) Run the Semi Definite Program (Optimisation)
% (IV)  Further details for unstable optimisation and mixed states







%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                              %
%       (I)  Importing the measured Probabilities              %
%                                                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%% We import a probability matrix with: 
%%% COLUMNS:  The events (0 clicks, 1 click, 2 cliks, etc). 
%%% ROWS:  The value of the corresponding input state. 
%%% (rows are the amplitude of |alpha|^2 if we are using 
%%% phase averaged coherent states to characterize diagonal POVMs)



%load('measured_probabilities.csv')';   %% this would import the data


%%% for demonstration purposes, we generate a theoretical Probability
%%% distribution corresponding to a lossy photon number detector:
%%% we generate it in file example_photon_number_detector_prob_distrib.m
%%% and include some statistical noise in the values of rho=|alpha><alpha|.
%%% (2% random noise added each |alpha| with a Gaussian distribution) 

PD=example_photon_number_detector_prob_distrib(25);   %% Probability Distribution
%%% (remember to add this folder to the MATLAB path)
        







%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                              %
%       (II)        Defining the INPUT State                   %
%                                                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




%%%%%%%%%%% DEFINE DIMENSIONS & PARAMETERS %%%%%%%%%%%%%%%%%%%

si = size(PD);                   %Determines size of the data set
imax = si(2);                    % Number of data points
mmax = floor(1+PD(1,si(2)));    % Determine cut-off in Fock space 



%%%  Pr_(alpha,n) = Tr(|alpha><alpha| POV_n) 
%%% will be expressed in matrix form:  P = F*POV.
%%% see page 3 in arXiv:0906.3440

%%%%%%%%%%%%%%%  LET US DEFINE THE "F"  MATRIX   %%%%%%%%%%%%%%%%%%

%%% F_{alpha,n} = exp(-|alpha|^2) |alpha|^{2n} / n!

F = zeros(mmax,imax);

for i1 = 1:mmax
        F(i1,1:imax)= exp(-(PD(1,1:imax).^2)).*(PD(1,1:imax).^(2*(i1-1)))./(factorial(i1-1)) ;
end

%% Warning! (this code may fail for large mmax, as factorial becomes Inf)
%% One would have to rewrite a more robust factorial.








%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%   SDP - Optimisation  %%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% In it's simplest version, we minimize the above relation:
%%% min_{POV}|| Pr - Tr(|alpha><alpha| POV) ||
%%% which in matrix form becomes, min_POV ||P - F*POV||
%%% since the Frobenius norm ||.|| is quadratic, POV>0 semi definite, 
%%% this is a quadratic optimisation problem
%%% see page 3 in arXiv:0906.3440


        %%%
    %%%%%%%%%%%%
%%%%% WARNING %%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%                                                                   %
%%% the min_POV ||P - F*POV|| optimisation is unstable if             %
%%% F is badly conditioned (as is the case with  |alpha><alpha|) and  %
%%% POV is a POVM from a lossy detector (having exponential tails for %
%%% higher photon numbers).                                           %
%%%                                                                   %
%%% If you want to regularize the optimsation, you will need some     %
%%% standard quadratic filter.  (see next section for details)        %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%% declare sdpvar (semi-definite-programming-variable) 
%%% & declare objective funtion (to be minimized)

pov = sdpvar(si(1)-1,mmax);  %% a 6 outcome POVM expanded up to fock layer mmax
Min_Object = (PD(2:si(1),:)-pov*F);  %% this is (P - F*POV)







%%%%%%%%%%%%%%%  simple (unstable) SDP Optimization %%%%%%%%%%%%%%%%%%%%%%
                                                                         %
%%% declare condition: POVM is positive:                                 %
%%% It's elementwise positive which means diagonal positive              %
%%% because of the specific representation we use.                       %  
CON = set(pov(:) > 0);                                                   %
%%% declare condition: \sum POVMS = identity                             %
CON = CON + set(sum(pov)==1);                                            %
%%% Solve the minimisation, subject to the above conditions, taking the  %
%%% Frobenius norm of (P-F*POV)                                          %
solvesdp(CON,norm(Min_Object,'fro'))  %%                                 %
                                                                         %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%                          

%%% RESULT:  note we must convert the SDP-variable back to a double
%%% precision matrix for us to be able to manipulate it in MATLAB.

POVM = double(pov); 





%%%%%%%%%%%  END OF SIMPLE PROGRAM:  %%%%%%%%%








%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                              %
%       (IV)      Instabilities and Mixed States               %
%                                                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%     To treat the unstable optimisation and consider mixed probe states (rho)
%%     see below


          %%%%%%%%%%%%%%%%%%%%%%%%%%
          % Mixed Coherent States  % 
          %%%%%%%%%%%%%%%%%%%%%%%%%%
          
          
%%% Instead of pure |alpha><alpha| your laser probably generates a mixture
%%% of coherent states (jitter in the amplitude, phase drift, etc).
%%% Thus, in our matrix "F" we need a description beyond the one used above.



%%%  for a phase averaged coherent state, 
%%%  with a gaussian amplitude distribution the probe state becomes:
%%%  rho = \int d\alpha \frac{1}{\sqrt{2\pi\sigma^2}}...
%%%  exp{-\frac{(\alpha-\alpha_0)^2}{2 \sigma^2}}

%%%  which expanded in the Fock basis can be written as,
%%%  F_{m,\alpha}=\sqrt{\frac{1}{2\pi\sigma^2}}...
%%%  \frac{1}{m!}exp(-\frac{\alpha^2}{1+2\sigma^2}...
%%%  \sum_{r=0}^{m} \binomial{2m,2r} ...
%%%  (\frac{2\sigma^2}{1+2\sigma^2})^{r+1/2}
%%%  (\frac{\sigma}{1+2\sigma^2})^{2m-2r} \Gamma(r+1/2).

%%% This can be generated with

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
% To speed up the code we create                              %
% binomial coefficients, factorials and gamma function        %
for i1 = 1:2*(mmax+1)                                         %
    bino(i1,1) = 1;                                           %
    for i2 = 2:i1
        m1=i1-1;
        m2=i2-1;
        bino(i1,i2) = bino(i1,i2-1)*(m1-m2+1)/m2;
    end
end
fac(1)=1;
for i1=2:2*mmax
    fac(i1)=fac(i1-1)*(i1-1);
end
for i1 = 0:mmax+1                                             %
    ga(i1+1)=gamma(i1+1/2);                                   %
end                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 



%%%%% we define the F matrix in |Pr-F*POV|, with mixed input states.
Fmix = zeros(mmax,imax);
for i1 = 1:mmax
    for i2 = 1:imax
        s = 0.02; % Standard deviation 
                  % (2% gaussian noise in this particular case)
        
        m = i1-1;
        for r = 0:m
            c1 = bino(2*m+1,2*r+1);
            c2 = (2*s*s/(1+2*s*s))^(r)*sqrt(2/(1+2*s*s));
            c3 = (sqrt(PD(1,i2))/(1+2*s*s))^(2*(m-r));
            Fmix(i1,i2) = Fmix(i1,i2) + c1*c2*c3*ga(r+1);
        end
        Fmix(i1,i2) = Fmix(i1,i2)*(sqrt(1/2/pi)/fac(i1))*exp(-PD(1,i2)/(1+2*s*s));
    end
end

%%%% see pages 13-14 of arXiv:0906.3440 for more information




          %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
          % Unstable Optimisation (REGULARISATION)  % 
          %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
          


          %% QUADRATIC FILTER TO REGULARIZE THE OPTIMISATION %%

          
povre = sdpvar(si(1)-1,mmax);  %% the regularized POVM (new sdp-variable)
Min_Object_re = (PD(2:si(1),:)-povre*Fmix);  %% this is (P - F*POV)
          
          
          %% We make a function out of our SDP-variable which
          %% imposes smoothness: the difference between
          %% pov(k,i) and pov(k,i+1) must remain small.
          %% see discussion p 16. of arXiv:0906.3440 for more details
smooth=0;
for k=1:si(1)-1
   for i=1:mmax-1
       smooth = smooth + (povre(k,i)-povre(k,i+1))^2;
   end
end


sm=0.1;  %%% Weight of the quadratic regularization (or penalty)

%%%%%%%  SDP Optimization %%%%%%%%%%%%%%%%%%%%%%%%%%
                                                   %
                                                   %
CON = set(povre(:) > 0); %positive povm           % 
CON = CON + set(sum(povre)==1); % adds to one     %
solvesdp(CON,norm(Min_Object_re,'fro')+sm*smooth)  %
                                                   %
%%%%%%%  SDP Optimization %%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%% RESULT  (converted from sdp-variable do double) %%

POVMre = double(povre);                          %
                            




%%%%%%%%%%%%  END OF CODE    %%%%%%%%%%%%%%
